/*------------------------------------------------------------------------------*/
/*	This source is entirely by IQ_132 -- http://www.freepgs.com/iq_132/	*/
/*	You may use this code in any way you wish, but you must include this	*/
/*	header with any files you distribute that contain any of this code	*/
/*------------------------------------------------------------------------------*/

#include <stdio.h>

// Taken from Final Burn Alpha source (src/burn/bitswap.h)
#define BITSWAP8(n,	\
	bit07, bit06, bit05, bit04, bit03, bit02, bit01, bit00)	\
	(((((n) >> (bit07)) & 1) <<  7) |	\
	 ((((n) >> (bit06)) & 1) <<  6) |	\
	 ((((n) >> (bit05)) & 1) <<  5) |	\
	 ((((n) >> (bit04)) & 1) <<  4) |	\
	 ((((n) >> (bit03)) & 1) <<  3) |	\
	 ((((n) >> (bit02)) & 1) <<  2) |	\
	 ((((n) >> (bit01)) & 1) <<  1) |	\
	 ((((n) >> (bit00)) & 1) <<  0))

void usage()
{
	printf ("Usage:\n");
	printf ("\tbswap8 input_file_1 input_file_2\n\n");
	printf ("\t - One of the input files must be the decrypted version of the other!\n\n");
}

// Write a Winhex-style diff. report
void report(unsigned char *sr1, unsigned char *sr2, int size, int size2, char *inp1, char *inp2)
{
	FILE *REPORT = fopen ("report.txt","wa");

	int i;
	int ctr = 0;
	for (i = 0; i < size; i++)
	{
		if (sr1[i] != sr2[i])
		{
			if (ctr == 0)
			{
				fprintf (REPORT, "Search for differences\n\n");
				fprintf (REPORT, "1. %s: %d bytes\n", inp1, size);
				fprintf (REPORT, "2. %s: %d bytes\n\n", inp2, size2);
				fprintf (REPORT, "Offsets: hexadec.\n\n");
			}
			ctr++;

			fprintf (REPORT, " %X:\t%2.2X\t%2.2X\n", i, sr1[i], sr2[i]);
		}
	}

	if (ctr != 0)
	{
		fprintf (REPORT, "\n%d difference(s) found. ", ctr);
		 printf ("However, there are %d different bytes in these files\n", ctr);
		 printf ("A diff. report (report.txt) has been written for you.\n");
	}
}

int main(int argc, char **argv)
{
	int a[8] = {0, 0, 0, 0, 0, 0, 0, 0}; // Store bitswap vars (8)
	int i, f, str = 0, cnt_total = 0; // Store other variables

	int b[8] = {0, 0, 0, 0, 0, 0, 0, 0}; // Store bitswap vars for "best match" (8)
	int b_mchs = 0; 		// Store number of matches for "best match"

	FILE *input_1, *input_2; // define encrypted and decrypted file.

	int size_1 = 0; // size of encrypted file
	int size_2 = 0; // size of decrypted file
	int size_3 = 0;

	if (argc < 3)
	{
		usage();
		return(1);
	}

	if ((input_1=fopen(argv[1],"rb"))==NULL)
	{
		fprintf(stderr,"Error: cannot read input_file_1 (%s).\n", argv[1]);
		return(1);
	}
	fseek (input_1 , 0 , SEEK_END);size_1 += ftell (input_1);rewind (input_1);

	if ((input_2=fopen(argv[2],"rb"))==NULL)
	{
		fprintf(stderr,"Error: cannot read input_file_2 (%s).\n", argv[2]);
		return(1);
	}
	fseek (input_2 , 0 , SEEK_END);size_2 += ftell (input_2);rewind (input_2);

	size_3 = size_1;

	if (size_1 != size_2)
	{
		printf ("The file sizes do not match.\n");

		if (size_1 > size_2)
		{
			size_1 = size_2;
		}
	}

	unsigned char *sr1 = (unsigned char*)malloc(size_1);
	unsigned char *sr2 = (unsigned char*)malloc(size_1);
	unsigned char *sr3 = (unsigned char*)malloc(size_1);

	fread (sr1,1,size_1,input_1); fclose(input_1);
	fread (sr2,1,size_1,input_2); fclose(input_2);

	for (a[0] = 0;  a[0] < 8; a[0]++) // var 1
	{
		for (a[1] = 0;  a[1] < 8; a[1]++) // var 2
		{
			if (a[1]!=a[0])
			for (a[2] = 0; a[2] < 8; a[2]++) // var 3
			{
				if ((a[2]!=a[1])&&(a[2]!=a[0]))
				for (a[3]= 0; a[3] < 8; a[3]++) // var 4
				{
					if ((a[3]!=a[2])&&(a[3]!=a[0])&&(a[3]!=a[1]))
					for (a[4] = 0; a[4] < 8; a[4]++) // var 5
					{
						if ((a[4]!=a[3])&&(a[4]!=a[2])&&(a[4]!=a[1])&&(a[4]!=a[0]))
						for (a[5] = 0; a[5] < 8; a[5]++) // var 6
						{
							if ((a[5]!=a[0])&&(a[5]!=a[1])&&(a[5]!=a[2])&&(a[5]!=a[3])&&(a[5]!=a[4]))
							for (a[6] = 0; a[6] < 8; a[6]++) // var 7
							{
								if ((a[6]!=a[0])&&(a[6]!=a[1])&&(a[6]!=a[2])&&(a[6]!=a[3])&&(a[6]!=a[4])&&(a[6]!=a[5]))
								for (a[7] = 0; a[7] < 8; a[7]++) // var 8
								{
									if ((a[7]!=a[0])&&(a[7]!=a[1])&&(a[7]!=a[2])&&(a[7]!=a[3])&&(a[7]!=a[4])&&(a[7]!=a[5])&&(a[7]!=a[6]))
									{
										cnt_total++;			// Running count of valid bitswaps tried

										for (i = 0; i < size_1; i++) 
										{
											sr3[i] = BITSWAP8(sr2[i], a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]);
											if (sr3[i] == sr1[i]) str++;
											if (sr3[i] != sr1[i])
											{
												i = size_1; // If a match isn't made, skip this combination: this speeds things up considerably.
											}											
										}

										if (str > b_mchs)
										{
											for (f = 0; f < 8; f++) b[f] = a[f]; // Store best combination
											b_mchs = str;			// Store number of matches
										}

										if (str == size_1)
										{
											if ((a[0]==7)&&(a[1]==6)&&(a[2]==5)&&(a[3]==4)&&(a[4]==3)&&(a[5]==2)&&(a[6]==1)&&(a[7]==0))
											{
												printf ("The input files appear to be identical.\n");
											} else {
												//printf("rom[i] = BITSWAP8(rom[i], %d, %d, %d, %d, %d, %d, %d, %d)\n", a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]);
												printf ("\nfor (i = 0; i < 0x%X; i++)\n{\n\trom[i] = BITSWAP8(rom[i], %d, %d, %d, %d, %d, %d, %d, %d)\n}\n\n", size_1, a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7]);										
												printf("Found after %d total checks\n",cnt_total);
												printf("%d Bytes checked that match\n", str);
											}

											free(sr1);free(sr2);free(sr3);
											return (1);

										} else {
											str = 0;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}

	if (b_mchs != 0)
	{
		if ((b[0]==7)&&(b[1]==6)&&(b[2]==5)&&(b[3]==4)&&(b[4]==3)&&(b[5]==2)&&(b[6]==1)&&(b[7]==0))
		{
			printf ("Neither of these files appear to be encrypted with bitswap8\n");

			report(sr1, sr2, size_3, size_2, argv[1], argv[2]);

			free(sr1);free(sr2);free(sr3);
			return (1);
		}

		int ctr_l = 0;
		for (i = 0; i < size_1; i++) 
		{
			sr3[i] = BITSWAP8(sr2[i], b[0], b[1], b[2], b[3], b[4], b[5], b[6], b[7]);
			if (sr1[i] == sr3[i]) ctr_l++;
		}

		printf ("A perfect match was not found. The best one was this:\n");
		printf ("\nfor (i = 0; i < 0x%X; i++)\n{\n\trom[i] = BITSWAP8(rom[i], %d, %d, %d, %d, %d, %d, %d, %d)\n}\n\n", size_1, b[0], b[1], b[2], b[3], b[4], b[5], b[6], b[7]);
		printf ("This bitswap matched %d out of %d bytes tried.\n", ctr_l, size_1);
	} else {
		printf ("There were no matches at all. This file may be encrypted in a different way than BITSWAP8\n");
	}

	free(sr1);
	free(sr2);
	free(sr3);
}
